#include <iostream>

#include "bsp.hpp"

////////////////////////////////////////////////////////////////

#define VERBOSE 0

namespace CoD
{
    ////////////////////////////////////////////////////////////////

    BSP::BSP(char *self, char *filename) :
        cod_version(0),
        cod_size(0)
    {
        #if VERBOSE
        std::cout << "Initialising file...\n";
        #endif

        std::string _self(self);
        path = _self.substr(0, _self.find_last_of("/\\") + 1);

        std::string _filename(filename);
        map = _filename.substr(_filename.find_last_of("/\\") + 1);

        // opening input file
        FILE *input;
        input = fopen(filename, "rb+");
        if (!input)
        {
            std::cerr << "ERROR: Could not open input bsp file '" << filename << "'.\n";
            return;
        }

        // reading input file
        fseek(input, 0, SEEK_END);
        data = Data(new char[ftell(input)], ftell(input));
        fseek(input, 0, SEEK_SET);

        if (fread(data.buffer, 1, data.size, input) != data.size)
        {
            std::cerr << "ERROR: Could not read bsp file '" << filename << "'.\n";
            delete[] data.buffer;
            data.buffer = NULL;
            fclose(input);
            return;
        }
        fclose(input);

        // checking file format
        if (data.size < 9 || data.buffer[0] != 'I' || data.buffer[1] != 'B' || data.buffer[2] != 'S' || data.buffer[3] != 'P')
        {
            std::cerr << "ERROR: The bsp's file format is unsupported.\n";
            delete[] data.buffer;
            data.buffer = NULL;
            return;
        }

        // initialising data access variable
        char *i = data.buffer;
        i += 4;

        // version check
        DoubleWord version;
        version.b[0] = *i; i++;
        version.b[1] = *i; i++;
        version.b[2] = *i; i++;
        version.b[3] = *i; i++;

        // check if this is CoD1, CoD2, ...
        switch (version.u)
        {
            case 59:
            {
                cod1();
                break;
            }
            case 4:
            {
                cod2();
                break;
            }
            default:
            {
                std::cerr << "ERROR: The bsp's file version is unsupported.\n";
                delete[] data.buffer;
                data.buffer = NULL;
                return;
            }
        }

        if (verbose)
        {
            std::cout << '\n';
        }

        ////////

        read();
    }

    BSP::~BSP()
    {
        delete[] data.buffer;
    }

    void BSP::cod1()
    {
        cod_version = 1;
        cod_size = 33;
    }

    void BSP::cod2()
    {
        cod_version = 2;
        cod_size = 39;

        lump[0] = &BSP::lump_textures;
        lump[4] = &BSP::lump_planes;
        lump[7] = &BSP::lump_texture_groups;
        lump[8] = &BSP::lump_vertexes;
        lump[9] = &BSP::lump_triangles;
        lump[37] = &BSP::lump_entities;
    }

    void BSP::read()
    {
        #if VERBOSE
        std::cout << "Initialising index...\n\n"
                  << "Lump ##:     Offset         Size   \n"
                  << "-----------------------------------\n";
        #endif

        i = data.buffer;
        i += 8;

        // read size and offset of each lump
        while (((size_t) (i + 7) - (size_t) data.buffer) < data.size)
        {
            DoubleWord size;
            size.b[0] = *i; i++;
            size.b[1] = *i; i++;
            size.b[2] = *i; i++;
            size.b[3] = *i; i++;

            DoubleWord offset;
            offset.b[0] = *i; i++;
            offset.b[1] = *i; i++;
            offset.b[2] = *i; i++;
            offset.b[3] = *i; i++;

            // as long as the end of the file is not too close...
            if ((offset.u + size.u) <= data.size)
            {
                index.push_back(Index(offset.u, size.u));

                #if VERBOSE
                std::cout.fill('0');
                std::cout << "Lump ";
                if ((index.size() - 1) < 10)
                {
                    std::cout << '0';
                }
                std::cout << std::dec << (index.size() - 1) << ":   0x";

                std::cout.width(8);
                std::cout << std::hex << offset.u << "    0x";

                std::cout.width(8);
                std::cout << std::hex << size.u << '\n';
                #endif
            }
            else
            {
                std::cerr << "WARNING: The bsp's file is too short, the lump's directory is pointing outside the file, starting from lump '" << index.size() << "'.\n";
                break;
            }

            // if we reached the number of lumps used by this CoD
            if (index.size() == cod_size)
            {
                break;
            }
        }

        #if VERBOSE
        std::cout << "\nTotal size: 0x";

        std::cout.width(8);
        std::cout << std::hex << data.size << "\n\n";
        #endif

        ////////

        // run a function for each lump
        unsigned int n = 0;
        std::vector<Index>::iterator it;
        for (it = index.begin(); it < index.end(); it++)
        {
            if ((*it).size && lump.find(n) != lump.end())
            {
                i = data.buffer + (*it).offset;
                size = data.buffer + (*it).offset + (*it).size;

                (*this.*lump[n])(); // run the lump function!
            }
            n++;
        }
    }

    void BSP::lump_textures()
    {
    }

    void BSP::lump_planes()
    {
        while ((i + 15) < size)
        {
            Plane plane;
            DoubleWord dword;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            plane.normal[0] = dword.i;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            plane.normal[1] = dword.i;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            plane.normal[2] = dword.i;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            plane.distance = dword.i;

            planes.push_back(plane);
        }
    }

    void BSP::lump_texture_groups()
    {
        while ((i + 15) < size)
        {
            TextureGroup texture_group;
            DoubleWord dword;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            texture_group.texture = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            texture_group.vertex_offset = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = dword.b[3] = 0x00;
            texture_group.vertex_length = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = dword.b[3] = 0x00;
            texture_group.triangle_length = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            texture_group.triangle_offset = dword.u;

            texture_groups.push_back(texture_group);
        }
    }

    void BSP::lump_vertexes()
    {
        while ((i + 67) < size)
        {
            Vertex vertex;

            DoubleWord dword;
            for (unsigned int n = 0; n < 3; n++)
            {
                dword.b[0] = *i; i++;
                dword.b[1] = *i; i++;
                dword.b[2] = *i; i++;
                dword.b[3] = *i; i++;
                vertex.position[n] = dword.f;
            }
            i += 15;

            dword.b[0] = *i; i++;
            dword.b[1] = dword.b[2] = dword.b[3] = 0x00;
            vertex.alpha = dword.u;

            i += 40;
            vertexes.push_back(vertex);
        }
    }

    void BSP::lump_triangles()
    {
        while ((i + 5) < size)
        {
            Triangle triangle;

            DoubleWord dword;
            for (unsigned int n = 0; n < 3; n++)
            {
                dword.b[0] = *i; i++;
                dword.b[1] = *i; i++;
                dword.b[2] = dword.b[3] = 0x00;
                triangle.vertex[n] = dword.u;
            }
            triangles.push_back(triangle);
        }
    }

    void BSP::lump_entities()
    {
    }

    ////////////////////////////////////////////////////////////////

    /*bool initialise_file(Data &data, char *filename)
    {
        std::cout << "Initialising file...\n";

        // opening input file
        FILE *input;
        input = fopen(filename, "rb+");
        if (!input)
        {
            std::cerr << "ERROR: Could not open input bsp file '" << filename << "'.\n";
            return false;
        }

        // reading input file
        fseek(input, 0, SEEK_END);
        data = Data(new char[ftell(input)], ftell(input));
        fseek(input, 0, SEEK_SET);

        if (fread(data.buffer, 1, data.size, input) != data.size)
        {
            std::cerr << "ERROR: Could not read bsp file '" << filename << "'.\n";
            delete[] data.buffer;
            fclose(input);
            return false;
        }
        fclose(input);

        // checking file format
        if (data.size < 9 || data.buffer[0] != 'I' || data.buffer[1] != 'B' || data.buffer[2] != 'S' || data.buffer[3] != 'P')
        {
            std::cerr << "ERROR: The bsp's file format is unsupported.\n";
            delete[] data.buffer;
            return false;
        }

        // initialising data access variable
        char *i = data.buffer;
        i += 4;

        // version check
        DoubleWord version;
        version.b[0] = *i; i++;
        version.b[1] = *i; i++;
        version.b[2] = *i; i++;
        version.b[3] = *i; i++;

        switch (version.u)
        {
            case 59:
            {
                cod_version = 1;
                break;
            }
            case 4:
            {
                cod_version = 2;
                break;
            }
            default:
            {
                std::cerr << "ERROR: The bsp's file version is unsupported.\n";
                delete[] data.buffer;
                return false;
            }
        }

        std::cout << '\n';
        return true;
    }

    ////////////////////////////////////////////////////////////////

    bool initialise_index(Data &data, std::vector<Index> &index)
    {
        std::cout << "Initialising index...\n\n";

        std::cout << "Lump ##:     Offset         Size   \n"
                  << "-----------------------------------\n";

        char *i = data.buffer;
        i += 8;

        while (((size_t) (i + 7) - (size_t) data.buffer) < data.size)
        {
            DoubleWord size;
            size.b[0] = *i; i++;
            size.b[1] = *i; i++;
            size.b[2] = *i; i++;
            size.b[3] = *i; i++;

            DoubleWord offset;
            offset.b[0] = *i; i++;
            offset.b[1] = *i; i++;
            offset.b[2] = *i; i++;
            offset.b[3] = *i; i++;

            if (data.size >= (offset.u + size.u))
            {
                index.push_back(Index(offset.u, size.u));

                std::cout.fill('0');
                std::cout << "Lump ";
                if ((index.size() - 1) < 10)
                {
                    std::cout << '0';
                }
                std::cout << std::dec << (index.size() - 1) << ":   0x";

                std::cout.width(8);
                std::cout << std::hex << offset.u << "    0x";

                std::cout.width(8);
                std::cout << std::hex << size.u << '\n';
            }
            else
            {
                std::cerr << "WARNING: The bsp's file is too short, the lump's directory is pointing outside the file, starting from lump '" << index.size() << "'.\n";
                break;
            }

            if (index.size() == cod_size[cod_version])
            {
                break;
            }
        }

        std::cout << "\nTotal size: 0x";

        std::cout.width(8);
        std::cout << std::hex << data.size;
        std::cout << '\n';

        std::cout << '\n';
        return true;
    }

    ////////////////////////////////////////////////////////////////

    void initialise_vertexes(Data &data, std::vector<Index> &index, std::vector<Vertex> &vertexes)
    {
        char *i = data.buffer + index[8].offset;
        char *max = data.buffer + index[8].offset + index[8].size;
        while ((i + 67) < max)
        {
            Vertex vertex;

            DoubleWord dword;
            for (unsigned int n = 0; n < 3; n++)
            {
                dword.b[0] = *i; i++;
                dword.b[1] = *i; i++;
                dword.b[2] = *i; i++;
                dword.b[3] = *i; i++;
                vertex.position[n] = dword.f;
            }
            i += 15;

            dword.b[0] = *i; i++;
            dword.b[1] = dword.b[2] = dword.b[3] = 0x00;
            vertex.alpha = dword.u;

            i += 40;
            vertexes.push_back(vertex);
        }
    }

    ////////////////////////////////////////////////////////////////

    void initialise_faces(Data &data, std::vector<Index> &index, std::vector<Face> &faces)
    {
        char *i = data.buffer + index[9].offset;
        char *max = data.buffer + index[9].offset + index[9].size;
        while ((i + 5) < max)
        {
            Face face;

            DoubleWord dword;
            for (unsigned int n = 0; n < 3; n++)
            {
                dword.b[0] = *i; i++;
                dword.b[1] = *i; i++;
                dword.b[2] = dword.b[3] = 0x00;
                face.vertex[n] = dword.u;
            }
            faces.push_back(face);
        }
    }

    ////////////////////////////////////////////////////////////////

    void initialise_groups(Data &data, std::vector<Index> &index, std::vector<Group> &groups)
    {
        char *i = data.buffer + index[7].offset;
        char *max = data.buffer + index[7].offset + index[7].size;
        while ((i + 15) < max)
        {
            Group group;
            DoubleWord dword;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            group.texture = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            group.vertex_offset = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = dword.b[3] = 0x00;
            group.vertex_length = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = dword.b[3] = 0x00;
            group.face_length = dword.u;

            dword.b[0] = *i; i++;
            dword.b[1] = *i; i++;
            dword.b[2] = *i; i++;
            dword.b[3] = *i; i++;
            group.face_offset = dword.u;

            groups.push_back(group);
        }
    }*/

    ////////////////////////////////////////////////////////////////
}

////////////////////////////////////////////////////////////////

