#ifndef WINDOW_H
#define WINDOW_H

#include <sstream>
#include <cmath>
#include <windows.h>

#include <sdl/SDL.h>
#include <sdl/SDL_opengl.h>

#define FRAME_MEMORY 100

///////////////////////////////////////////////////////////////////////////////

namespace Window
{
    double current;
    char *title;

    // frame buffers to more accurately provide a stable fps and spf
    double fps_frame_memory[FRAME_MEMORY];
    double spf_frame_memory[FRAME_MEMORY];

    unsigned int frame_i = 0;
    bool one_loop = false;

    double GetSystemTime()
    {
        static LARGE_INTEGER Frequency;
        static BOOL UseHighPerformanceTimer = QueryPerformanceFrequency(&Frequency);

        if (UseHighPerformanceTimer)
        {
            LARGE_INTEGER CurrentTime;
            QueryPerformanceCounter(&CurrentTime);

            return static_cast<double>(CurrentTime.QuadPart) / Frequency.QuadPart;
        }
        else
        {
            return GetTickCount() * 0.001;
        }
    }

    double current_fps()
    {
        double time = (GetSystemTime() - current);
        if (time)
        {
            return (1 / time);
        }
        return 0;
    }

    double stable_fps()
    {
        double _fps = 0.0;
        if (one_loop)
        {
            for (unsigned int n = 0; n < FRAME_MEMORY; n++)
            {
                _fps += fps_frame_memory[n];
            }
            return (_fps / FRAME_MEMORY);
        }
        else if (frame_i)
        {
            for (unsigned int n = 0; n < frame_i; n++)
            {
                _fps += fps_frame_memory[n];
            }
            return (_fps / frame_i);
        }
        else
        {
            return 75.0;
        }

    }

    double current_spf()
    {
        return (GetSystemTime() - current);
    }

    double stable_spf()
    {
        double _spf = 0.0;
        if (one_loop)
        {
            for (unsigned n = 0; n < FRAME_MEMORY; n++)
            {
                _spf += spf_frame_memory[n];
            }
            return (_spf / FRAME_MEMORY);
        }
        else if (frame_i)
        {
            for (unsigned n = 0; n < frame_i; n++)
            {
                _spf += spf_frame_memory[n];
            }
            return (_spf / frame_i);
        }
        else
        {
            return 0.01333;
        }
    }

    void caption_fps()
    {
        std::stringstream caption;
        caption << title << " - FPS: " << stable_fps() << " SPF: " << stable_spf();
        SDL_WM_SetCaption(caption.str().c_str(), NULL);
    }

    void initialise(char *ititle, int width, int height, bool fullscreen, int depth)
    {
        title = ititle;

        if (SDL_Init(SDL_INIT_VIDEO) < 0)
        {
            exit(1);
        }

        SDL_GL_SetAttribute(SDL_GL_RED_SIZE,   8);
        SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
        SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,  8);
        SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,  8);
        SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 32);
        SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
        //SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, 0);

        SDL_WM_SetCaption(title, NULL);

        const SDL_VideoInfo *video_info = SDL_GetVideoInfo();
        if (!video_info)
        {
            SDL_Quit();
            exit(1);
        }

        int flags = SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE;

        if (video_info->hw_available)
        { flags |= SDL_HWSURFACE; }
        else
        { flags |= SDL_SWSURFACE; }

        if (video_info->blit_hw)
        { flags |= SDL_HWACCEL; }

        if (fullscreen)
        { flags |= SDL_FULLSCREEN; }

        SDL_Surface *screen = SDL_SetVideoMode(width, height, depth, flags);
        if (!screen)
        {
            SDL_Quit();
            exit(1);
        }

        SDL_EnableUNICODE(SDL_ENABLE);

        current = GetSystemTime();

        for (unsigned int i = 0; i < FRAME_MEMORY; i++)
        {
            fps_frame_memory[i] = 75;
            spf_frame_memory[i] = 0.01333;
        }
    }

    ///////////////////////////////////////////////////////////////////////////////

    void pre_frame()
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    void post_frame()
    {
        SDL_GL_SwapBuffers();

        fps_frame_memory[frame_i] = current_fps();
        spf_frame_memory[frame_i] = current_spf();

        frame_i++;
        if (frame_i == FRAME_MEMORY)
        {
            one_loop = true;
            frame_i = 0;
        }

        current = GetSystemTime();
    }

    void delay()
    {
        SDL_Delay(1);
    }

    ///////////////////////////////////////////////////////////////////////////////

    void shutdown()
    {
        SDL_Quit();
    }
}

///////////////////////////////////////////////////////////////////////////////

#endif // WINDOW_H
